"""
BlackJack Simulator

A Python library for playing and simulating Black Jack games with customizable strategies.
"""

__version__ = "0.7.0"
__author__ = "Michael Munson"
__email__ = "michaelmunsonm@gmail.com"

# Import main classes for easy access
from .game import Game, Simulation
from .player import Player, Dealer, Strategy, Simple, Simple17
from .deck import Deck, Hand, Card

# Import constants
from .player import HIT, STAY, INSURANCE, DOUBLE_DOWN, SPLIT

__all__ = [
    # Game classes
    "Game",
    "Simulation",
    
    # Player classes
    "Player",
    "Dealer",
    "Strategy",
    "Simple",
    "Simple17",
    
    # Deck classes
    "Deck",
    "Hand", 
    "Card",
    
    # Constants
    "HIT",
    "STAY",
    "INSURANCE",
    "DOUBLE_DOWN",
    "SPLIT",
]



#!/usr/bin/env python3
"""
Command-line interface for the BlackJack simulator.
"""

import argparse
import sys
from .game import Simulation, Game
from .player import Player, Simple, Simple17

def main():
    parser = argparse.ArgumentParser(
        description="BlackJack Simulator - Run simulations with different strategies"
    )
    parser.add_argument(
        "--players", 
        nargs="+", 
        default=["Player1", "Player2"],
        help="Player names (default: Player1 Player2)"
    )
    parser.add_argument(
        "--chips", 
        type=int, 
        default=1000,
        help="Starting chips per player (default: 1000)"
    )
    parser.add_argument(
        "--rounds", 
        type=int, 
        default=1000,
        help="Number of rounds to simulate (default: 1000)"
    )
    parser.add_argument(
        "--min-bet", 
        type=int, 
        default=15,
        help="Minimum bet (default: 15)"
    )
    parser.add_argument(
        "--strategy", 
        choices=["simple", "simple17"], 
        default="simple",
        help="Strategy to use (default: simple)"
    )
    parser.add_argument(
        "--verbose", 
        action="store_true",
        help="Show simulation progress"
    )
    parser.add_argument(
        "--interactive", 
        action="store_true",
        help="Play a single interactive game instead of running simulations"
    )

    args = parser.parse_args()

    # Create strategy
    strategy_map = {
        "simple": Simple(),
        "simple17": Simple17(),
    }
    strategy = strategy_map[args.strategy]

    # Create players
    players = [
        Player(name, chips=args.chips, strategy=strategy)
        for name in args.players
    ]

    if args.interactive:
        # Play interactive game
        game = Game(players=players, min_bet=args.min_bet)
        game.start()
    else:
        # Run simulation
        sim = Simulation(players=players, min_bet=args.min_bet)
        results = sim.run(n_times=args.rounds, print_sim=args.verbose)
        
        # Print results
        results.print()

if __name__ == "__main__":
    main() 


import pytest
from .game import Game, Simulation
from blackjack.player import Player, Simple
from blackjack.deck import Deck

def test_game_creation():
    players = [Player("Test", strategy=Simple())]
    game = Game(players=players)
    assert len(game.players) == 1
    assert game.min_bet == 15

def test_simulation():
    players = [Player("Test", strategy=Simple())]
    sim = Simulation(players=players)
    results = sim.run(n_times=10)
    assert len(results) == 1
    assert "Test" in results 


from blackjack.game import Game, Simple, Player, Dealer, Simulation
from blackjack.deck import Deck, Card, Hand
from blackjack.player import Simple, Simple17, PseudoPlayer, Strategy, INSURANCE, SPLIT, STAY, HIT, DOUBLE_DOWN
from random import randint
from escprint import esc
from typing import TypedDict, NamedTuple
import typing

class MoreComplex(Strategy):
    # randomly decide between 1 and 2 hands
    def decide_hands(self, player):
        # include these arguments, even if they're not used
        return randint(1,2)
    # if the player is net positive on chips, bet twice the amount
    def decide_bet(self, player, min_bet):
        # include these arguments, even if they're not used
        if player.chips > player.init_chips:
            return int(2 * min_bet)
    #
    def decide(self, player, choices, dealer, players):
        # include these arguments, even if they're not used
        player_val = player.hand_value()

        if INSURANCE in choices:
            return INSURANCE
        
        elif dealer.showing() in [4,5,6]:
            return STAY
        
        elif SPLIT in choices:
            return SPLIT

        elif DOUBLE_DOWN in choices and player_val <= 13 and player_val >= 11:
            return DOUBLE_DOWN

        elif player_val < dealer.showing() + 10 and player_val < 16:
            return HIT

class CardCounting(Strategy):
    # initialize state field
    def init_state(self) -> None:
        self.state = {
            "Aces" : 0
        }
    # after round is over
    def after(self, player, dealer, players):
        if "Ace" in map(lambda card: card.rank, player.hand):
            self.state["Aces"] += 1 
        
        for p in players:
            if "Ace" in map(lambda card: card.rank, p.hand):
                self.state["Aces"] += 1

        if "Ace" in map(lambda card: card.rank, dealer.hand):
            self.state["Aces"] += 1 

print_tname = esc.create_fn("White")
def _print_succ():
    esc.print("✔︎ All Tests Passed", "Green")
def _print_err():
    esc.print("𐄂 Test Failed", "red")

## ACTUAL TESTS

def test_hand():
    esc.print("*** Test Hand Methods ***", "White")
    try: 
        hand = Hand(Card("Ace","hearts"), Card("9", "diamonds"), Card("8","spades"))
        assert hand.value() == 18, "hand.value() == 18"
        hand = Hand(Card("Ace","spades"), Card("Ace","diamonds"), Card("Ace", "hearts"), Card("2", "hearts"))
        assert hand.value() == 15, "hand.value() == 15"
        assert hand.lowest_value() == 5, "hand.low_value() == 5"
        _print_succ()
    except AssertionError as error:
        _print_err()

def test_is_blackjack():
    print_tname("*** TEST Hand.is_blackjack")
    try:
        hand = Hand(Card("Ace"), Card("King"))
        assert hand.is_blackjack(), "Hand (Ace,King) should return blackjack"
        hand = Hand("King","9")
        assert not hand.is_blackjack(), "Hand (King,9) should return NOT blackjack"
        _print_succ()
    except AssertionError as error:
        _print_err()
        print(error)

def test_input_bet():
    print_tname("*** Test Input Bet ***")
    init_chips = 1000
    player = Player("Mike", chips=init_chips)
    bet = player.input_bet(min_bet=15)
    assert player.chips == (init_chips - bet), "Discrepancy between difference between initial chips and bet amount, and actual result."
    _print_succ()

def test_split_hand():
    print_tname("*** TEST Player.split_hand ***")
    player = Player("Mike", 1000)
    player.bet = 100
    player.hand = Hand("5","5")
    player.split_hand()
    assert player.pseudos[0].hand.len() == 1 and player.pseudos[1].hand.len() == 1, "Hand length should be 1"
    assert player.pseudos[0].hand[0].value == 5 and player.pseudos[1].hand[0].value == 5, "Hand values should be 5"
    assert player.bet == 200, "Player bet should be 200"
    _print_succ()

def test_pseudo_place_bet():
    print_tname("*** TEST PseudoPlayer.place_bet ***")
    player = Player("Mike",chips=1000)
    pseudo = PseudoPlayer("Mike-pseudo",parent=player, bet=15)
    # print(player.chips)
    # print(f"bet: {pseudo.bet}")
    pseudo.place_bet(15)
    # print(f"bet: {pseudo.bet}")
    # print(f"player chips = {player.chips}")
    assert pseudo.bet == 30, "Pseudo bet should be 30"
    assert player.chips == 970, "Player chips should be 970"
    assert player.bet == 30, "Player bet should be 30"
    _print_succ()

# OTHER
def test_game():
    players = [
        Player("Mike", chips=1000),
        Player("Dan", chips=1000)
    ]

    dealer = Dealer()

    deck = Deck(shuffle=True, num_decks=8)

    game = Game(
        players=players, 
        dealer=dealer,
        deck=deck, 
        min_bet=15,
        hit_on_soft_17=True
    )

    game.start()

def test_print():
    player = Player("Mike")
    player.hit(Card("Ace"))
    player.hit(Card("2"))
    player.print()

def test_start():
    players = [
        Player("Mike", chips=1000),
        # Player("Dan", chips=1000),
        # Player("John", chips=1000)
    ]

    game = Game(players=players, min_bet=15)

    game.start()

def test_sim_start():
    sim = Simulation(
        players=[
            Player("Mike",1000)
        ],
        min_bet=15
    )

    sim._start(is_print=True)

def test_sim_run(n_times:int=1, print_sim:bool=True, wait:float=.01):
    ccstrat = CardCounting()
    sim = Simulation(
        players=[
            Player("Mike",10000, strategy=MoreComplex()),
            Player("Cheater",10000, strategy=ccstrat)
        ],
        min_bet=15
    )

    sim_results = sim.run(n_times=n_times, print_sim=print_sim, wait=wait)
    
    sim_results.print()

    print(ccstrat.state)
    # for player in sim_results.players:
    #     print(player.chips - player.init_chips)

    # print(sum(res.won for res in mike.results))


if __name__ == "__main__":
    # test_is_blackjack()
    # test_split_hand()
    # test_game()
    # player = Player("Mike", 100)
    # pseudo = PseudoPlayer("Mike1",parent=player, bet=player.bet)
    # pseudo.hit(Card("5"))
    # print(pseudo.hand)
    # test_input_bet()
    # test_pseudo_place_bet()
    # test_split_hand()
    # test_sim_start()
    # test_sim_run(n_times=10000, print_sim=False, wait=.1)

    test_sim_run(n_times=1000, print_sim=False, wait=.01)
    # hand = Hand("King","King")
    # print(hand)
    # if "Ace" in map(lambda card: card.rank, hand):
    #     print(True)



from __future__ import annotations
import random
from typing import NamedTuple


class Card:
    SYM_MAP = {
        "hearts" : "❤️",
        "spades" : "♠︎",
        "diamonds" : "♦️",
        "clubs" : "♣︎"
    }

    def __init__(self, rank:str, suit:str="spades") -> None:
        if not Card.is_rank(rank):
            raise TypeError(f"'{rank}' not a valid card rank.")
        if not Card.is_suit(suit):
            raise TypeError(f"'{suit}' not a valid card suit.")

        self.rank = str.upper(rank[0]) + str.lower(rank[1:])
        self.suit = self.SYM_MAP[suit.lower()] if suit.lower() in self.SYM_MAP else suit
        if rank.isdigit():
            self.value = int(rank)
        elif rank == "Ace":
            self.value = 11
        else:
            self.value = 10
    
    def to_str(self) -> str:
        return f"{self.rank} {self.suit}"

    @staticmethod
    def is_card(card) -> bool:
        return isinstance(card,Card)

    @staticmethod
    def is_rank(rank:str) -> bool:
        return rank in ['Ace', '2', '3', '4', '5', '6', '7', '8', '9', '10', 'Jack', 'Queen', 'King']

    @staticmethod
    def is_suit(suit:str) -> bool:
        return str.lower(suit) in ['hearts','diamonds','clubs','spades', '❤️', '♦️', '♣︎', '♠︎'] 

class Hand(list[Card]):
    bet:int
    
    def __init__(self, *cards:Card|str, bet:int=0):
        self.bet = bet
        for card in cards:
            self.add(card)

    def add(self,card:Card|str) -> None:
        if Card.is_card(card):
            return self.append(card)
        elif isinstance(card,str):
            card = Card(card)
            return self.append(card)
        
    def value(self) -> int:
        value = 0
        for card in self:
            value += card.value

        if value > 21:
            i = 0
            while i < len(self) and value > 21:
                card = self[i]
                if card.rank == "Ace":
                    value -= 10
                i += 1

        return value
    
    def is_bust(self) -> int:
        return self.value() > 21
    
    def is_blackjack(self) -> int:
        return len(self) == 2 and self.value() == 21

    def lowest_value(self) -> int:
        value = 0
        for card in self:
            if card.rank == "Ace":
                value += 1
            else:
                value += card.value
        return value

    def to_str() -> str:
        return " | ".join(list(map(lambda card: card.to_str())))

    def len(self) -> int:
        return len(self)
    
    def split(self) -> tuple[Hand,Hand]:
        return (
            Hand(self[0]),
            Hand(self[1])
        )

class Deck(list):
    def __init__(self, num_decks:int=1, shuffle:bool=False):
        self.num_decks = num_decks
        self.is_shuffle = shuffle
        self.reset()
    # shuffle deck
    def shuffle(self):
        random.shuffle(self)
    # deal card
    def deal(self):
        if len(self) == 0:
            return None
        return self.pop()
    # reset deck
    def reset(self):
        self.clear()
        for i in range(self.num_decks):
            self += [Card(rank, suit)
                    for rank in ['Ace', '2', '3', '4', '5', '6', '7', '8', '9', '10', 'Jack', 'Queen', 'King']
                    for suit in ['❤️', '♦️', '♣︎', '♠︎']]
                    # for suit in ['Hearts', 'Diamonds', 'Clubs', 'Spades']]
        if self.shuffle:
            self.shuffle()
            self.shuffle()

    def card_str_list(self):
        return list(map(lambda card: card.to_str(), self))
    
    def peek(self,num_cards:int):
        return list(map(lambda card: card.to_str(), self[-num_cards:]))
    


from __future__ import annotations
from .deck import Card, Hand
from escprint import esc
from typing import NamedTuple

HIT = "hit"
STAY = "stay"
INSURANCE = "insurance"
DOUBLE_DOWN = "double down"
SPLIT = "split"

### STRATEGY ###
class Strategy:
    log:list
    auto_log:bool
    state_log: list
    state: dict

    def __init__(self, auto_log:bool=True) -> None:
        self.auto_log = auto_log
        self.log = []
        self.state_log = []
        self.state = {}
        self.init_state()

    def init_state(self):
        self.state = {}

    def __decide_hands__(self, player:Player) -> int:
        n_hands = self.decide_hands(player=player)
        if not n_hands:
            n_hands = 1
        if self.auto_log:
            self.log.append(f"hands -> {n_hands}")
        if not isinstance(n_hands,int):
            raise TypeError("Strategy decision for number of hands must be integer.")
        if n_hands < 1:
            raise ValueError("Strategy decision for number of hands must exceed 0")
        
        return n_hands
        
    def __decide_bet__(self, player:Player, min_bet:int=15) -> int:
        bet = self.decide_bet(player=player, min_bet=min_bet)
        if bet == None:
            bet = min_bet
        if not isinstance(bet, int):
            raise TypeError(f"Strategy bet decision must be integer")
        if bet < min_bet:
            raise ValueError(f"Strategy bet decision must be integer greater than min_bet ({min_bet})")
        if self.auto_log:
            self.log.append(f"bet -> {bet}")
        return bet
        
    def __decide__(self, player:Player, choices:list[str], dealer:Dealer=None, players:list[Player]=[]) -> str:
        decision = self.decide(
            player=player,
            choices=choices,
            dealer=dealer,
            players=players
        )

        if not decision:
            decision = STAY

        decision = str.lower(decision)

        if not isinstance(decision, str):
            raise TypeError("Strategy decision must be of type str")
        if decision not in choices:
            raise ValueError(f"Straregy decision not in list of valid choices: {choices}")
        
        return decision
        
    def __after__(self, player:Player, dealer:Dealer=None, players:list[Player]=[]) -> None:
        return self.after(player=player, dealer=dealer, players=players)

    def _reset_log(self) -> None:
        self.log.clear()

    def decide_hands(self, player:Player) -> int:
        return 1
    
    def decide_bet(self, player:Player, min_bet:int=15) -> int:
        return min_bet

    def decide(self, player:Player, choices:list[str], dealer:Dealer=None, players:list[Player]=[]) -> str:
        # possible choices = ["insurance","split","hit","stay","double down"]
        return STAY

    def after(self, player:Player, dealer:Dealer=None, players:list[Player]=[]) -> None:
        pass

class Simple(Strategy):
    def init_state(self):
        self.state = {
            "Aces" : 0
        }

    def decide_hands(self, player: Player) -> int:
        return 1
    
    def decide_bet(self, player: Player, min_bet: int = 15) -> int:
        return min_bet

    def decide(self, player:Player, choices:list[str], dealer:Dealer=None, players:list[Player]=[]) -> str:
        return HIT if player.hand_value() < 16 else STAY

class Simple17(Strategy):
    def decide(self, player:Player, choices:list[str], dealer:Dealer=None, players:list[Player]=[]) -> str:
        return HIT if player.hand_value() < 17 else STAY

### PLAYER ###
class Player: 
    name:str
    strategy:Strategy
    hand:Hand
    init_chips:int
    chips:int
    bet:int
    is_split:bool
    insurance:int
    pseudos:list[PseudoPlayer]
    is_pseudo:bool
    is_stayed:bool
    results:list[PlayerResults]

    def __init__(self, name:str, chips:int=1000, strategy:Strategy=Simple()) -> None:
        self.name = name
        self.strategy = strategy
        self.hand = Hand()
        self.init_chips = chips
        self.chips = chips
        self.bet = 0
        self.insurance = 0
        self.pseudos = []
        self.is_pseudo = False
        self.is_stayed = False
        self.results = []
    
    def hit(self, card:Card) -> bool:
        self.hand.add(card)
        return self.is_bust()
    
    def is_bust(self) -> bool:
        return self.hand.is_bust()

    def has_blackjack(self) -> bool:
        return self.hand.is_blackjack()
    
    def can_split(self) -> bool:
        return self.hand.len() == 2 and self.hand[0].rank == self.hand[1].rank

    def hand_value(self) -> int:
        return self.hand.value()
    
    def lowest_hand_value(self) -> int:
        return self.hand.lowest_value()

    def card_str_list(self) -> list:
        return list(map(lambda card: card.to_str(), self.hand))

    def print(self, max_name_len:int=0, dealer:Dealer=None) -> None:
        dealer_hand_value = -1 if dealer == None else dealer.hand_value()

        card_str_arr = []
        for card in self.hand:
            card_str_arr.append(f"{card.to_str()}")

        pref_len = max_name_len - len(self.name)
        post_str = ""
        if pref_len > 0:
            post_str = " " * pref_len
        
        print_style = "red" if (self.is_bust() or (dealer_hand_value > self.hand_value() and dealer_hand_value < 22) or (dealer_hand_value == 21 and len(dealer.hand) == 2)) else "Green/bold"
        print_strike = "strike" if (self.is_bust() or (dealer_hand_value > self.hand_value() and dealer_hand_value < 22) or (dealer_hand_value == 21 and len(dealer.hand) == 2)) else ""
        print_hand_style = "red" if (self.is_bust() or (dealer_hand_value > self.hand_value() and dealer_hand_value < 22) or (dealer_hand_value == 21 and len(dealer.hand) == 2)) else "Cyan/bold"
        print_red = "red" if (self.is_bust() or (dealer_hand_value > self.hand_value() and dealer_hand_value < 22) or (dealer_hand_value == 21 and len(dealer.hand) == 2)) else ""
        
        bet_str = f"${self.bet}" if not self.has_insurance() else f"${self.bet} + (${self.insurance})"
        stayed_check = " ✔︎" if self.is_stayed else ""

        if not self.has_pseudos():
            esc.printf(
                (self.name, print_style, print_strike, "underline"), (post_str,print_style), 
                (" ... ",print_red), (bet_str, print_style, "underline"),
                (" -> ",print_red), # (f"{post_str + (' '*len(self.name))} ... ", print_style), 
                (f"{' | '.join(card_str_arr)}", print_hand_style, print_strike),
                stayed_check
            )
        else:
            print_style = "red" if (self.is_bust() or (dealer_hand_value == 21 and len(dealer.hand) == 2)) else "Green/bold"
            print_strike = "strike" if (self.is_bust() or (dealer_hand_value == 21 and len(dealer.hand) == 2)) else ""

            esc.printf(
                (self.name, print_style, print_strike, "underline"), 
                (post_str,print_style)," ... ", (bet_str, print_style, "underline/dim"),
            )
            for pseudo in self.pseudos:
                pseudo.print(max_name_len=max_name_len, dealer=dealer)

    def reset(self) -> None:
        self.hand.clear()
        self.bet = 0
        self.insurance = 0
        self.is_stayed = False
        self.pseudos = []

    def place_bet(self, bet_amount:int, min_bet:int=15) -> int:
        if self.chips < bet_amount:
            return -2
        
        if min_bet > bet_amount:
            return -1
        
        self.bet += bet_amount
        self.chips -= bet_amount

        return self.bet

    def place_insurance_bet(self) -> None:
        bet = (self.bet / 2)
        self.insurance = bet
        self.chips -= bet

    def get_init_round_inputs(self, min_bet:int=15) -> None:
        esc.printf(
            f"{self.name}, How many hands? ", ("default = 1", "dim")  
        )

        hand_amount = (
            esc.input("#", input="Green", end="")
        )
        
        if not hand_amount.isdigit() and hand_amount != "":
            esc.erase_screen(); esc.cursor_to_top()
            esc.print("Hand amount must be integer.","Red/italic")
            return self.get_init_round_inputs(min_bet=min_bet)

        if hand_amount == "":
            hand_amount = 1
        else:
            hand_amount = int(hand_amount)

        esc.printf(
            f"{self.name} (", [f"${self.chips}", "Green/underline"], 
            f") What is your bet? ", (f"defualt = ${min_bet}","dim")  
        )

        bet_amount = (
            esc.input("$", input="Green", end="")
        )
                
        if not bet_amount.isdigit() and bet_amount != "":
            esc.erase_screen(); esc.cursor_to_top()
            esc.print("Bet amount must be integer.","Red/italic")
            return self.get_init_round_inputs(min_bet=min_bet)

        if bet_amount == "":
            bet_amount = min_bet
        else:
            bet_amount = int(bet_amount)


        if (hand_amount * bet_amount > self.chips):
            esc.erase_screen(); esc.cursor_to_top()
            esc.print("Bet * Hand Amt. greater than chip count","Red/italic")
            return self.get_init_round_inputs(min_bet=min_bet)
        if (bet_amount < min_bet):
            esc.erase_screen(); esc.cursor_to_top()
            esc.print("Bet amount lower than Min Bet.","Red/italic")
            return self.get_init_round_inputs(min_bet=min_bet)

        if hand_amount > 1:
            for i in range(hand_amount):
                self.pseudos.append(
                    PseudoPlayer(name=f"{self.name}", parent=self, bet=bet_amount)
                )
        else:
            self.place_bet(bet_amount=bet_amount, min_bet=min_bet)

    def has_pseudos(self) -> bool:
        return len(self.pseudos) > 0

    def has_insurance(self) -> bool:
        return self.insurance > 0

    def split_hand(self) -> None:
        # weed out bad calls
        if not self.can_split():
            return
        # rest
        (hand1,hand2) = self.hand.split()

        bet = self.bet
        self.bet = 0

        self.pseudos = [
            PseudoPlayer(self.name, parent=self, bet=bet, hand=hand1),
            PseudoPlayer(self.name, parent=self, bet=bet, hand=hand2)
        ]
        #
        # self.bet *= 2

    def _handle_mult_hands(self, hand_amount:int, bet_amount:int) -> None:
        if hand_amount > 1:
            for i in range(hand_amount):
                self.pseudos.append(
                    PseudoPlayer(name=f"{self.name}", parent=self, bet=bet_amount)
                )

### Pseudo PLAYER ###
class PseudoPlayer(Player):
    def __init__(self, name:str, parent:Player, bet:int, hand:Hand=None) -> None: # hand = Hand() causes weird bug
        super().__init__(name)
        self.parent = parent
        self.hand = hand if hand != None else Hand() # again have to do cause weird bug
        self.is_pseudo = True
        self.place_bet(bet_amount=bet)
    
    def print(self, max_name_len:int=0, dealer:Dealer=None) -> None:
        dealer_hand_value = -1 if dealer == None else dealer.hand_value()

        card_str_arr = []
        for card in self.hand:
            card_str_arr.append(f"{card.to_str()}")

        pref_len = max_name_len - len(self.name)
        post_str = ""
        if pref_len > 0:
            post_str = " " * pref_len
        
        print_style = "red" if (self.is_bust() or (dealer_hand_value > self.hand_value() and dealer_hand_value < 22) or (dealer_hand_value == 21 and len(dealer.hand) == 2)) else "Green/bold"
        print_strike = "strike" if (self.is_bust() or (dealer_hand_value > self.hand_value() and dealer_hand_value < 22) or (dealer_hand_value == 21 and len(dealer.hand) == 2)) else ""
        print_red = "red" if (self.is_bust() or (dealer_hand_value > self.hand_value() and dealer_hand_value < 22) or (dealer_hand_value == 21 and len(dealer.hand) == 2)) else ""
        print_hand_style = "red" if (self.is_bust() or (dealer_hand_value > self.hand_value() and dealer_hand_value < 22) or (dealer_hand_value == 21 and len(dealer.hand) == 2)) else "Cyan/bold"
        stayed_check = " ✔︎" if self.is_stayed else ""

        esc.printf(
            (f"{post_str + (' '*len(self.name))} ... ${self.bet}", print_style, print_strike), 
            (" -> ",print_red), (f"{' | '.join(card_str_arr)}", print_hand_style, print_strike),
            stayed_check
        )

    def place_bet(self, bet_amount: int, min_bet: int = 15) -> int:
        if self.parent.chips < bet_amount:
            return -2
        
        if min_bet > bet_amount:
            return -1
        
        self.bet += bet_amount
        self.parent.bet += bet_amount
        self.parent.chips -= bet_amount

        return self.bet

    def split_hand(self):
        bet = self.bet
        self.bet = 0
        # weed out bad calls
        if not self.can_split():
            return
        # rest
        (hand1,hand2) = self.hand.split()
        index = self.parent.pseudos.index(self)
        self.parent.pseudos.pop(index)
        self.parent.pseudos.insert(index, PseudoPlayer(self.name, parent=self.parent, bet=bet, hand=hand1))
        self.parent.pseudos.insert(index+1,  PseudoPlayer(self.name, parent=self.parent, bet=bet, hand=hand2))
        # self.parent.bet += self.bet

class Dealer(Player):
    def __init__(self, strategy:Strategy=Simple17()) -> None:
        super().__init__("Dealer", strategy)

    def print(self, hidden=True, max_name_len:int=0) -> None:
        card_str_arr = []
        for i in range(len(self.hand)):
            if hidden and i > 0:
                card_str_arr.append("*******")
            else:
                card_str_arr.append(f"{self.hand[i].to_str()}")

        pref_len = max_name_len - len(self.name)
        post_str = ""
        if pref_len > 0:
            for _ in range(pref_len):
                post_str+= " "

        esc.printf(
            (self.name, "red/strikethrough" if self.is_bust() else "Blue/bold/underline"), post_str,
            (f" ... {' | '.join(card_str_arr)}", "red/strikethrough" if self.is_bust() else "Cyan/bold")
        )

    def showing(self) -> int:
        if len(self.hand) > 0:
            return self.hand[0].value
        return 0
    
    def showing_ace(self) -> bool:
        return self.hand[0].rank == "Ace"
    
### PLAYER RESULTS
class PlayerResults(NamedTuple):
    player:Player
    hands:int
    won:int
    pushed:int
    busted:int
    net:int

class PlayerSimulationResults(NamedTuple):
    player:Player
    rounds:int
    hands:int
    won:int
    pushed:int
    busted:int
    net:int
    win_rate:float
    


from setuptools import setup, find_packages

setup(
    name="jackblack",
    version="0.7.0",
    author="Michael Munson",
    author_email="michaelmunsonm@gmail.com",
    description="A Blackjack simulator and strategy engine",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/michaelmunson/blackjack",
    packages=find_packages(),  # auto-discovers `jackblack`
    include_package_data=True,
    install_requires=[],  # or read from requirements.txt
    entry_points={
        "console_scripts": [
            "jackblack=jackblack.cli:main",  # if you want a CLI
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
)



from __future__ import annotations
from .deck import Deck, Hand, Card
from .player import Player, Dealer, Simple, PlayerResults, PlayerSimulationResults
from escprint import esc
from collections import namedtuple
from time import sleep, time
from typing import NamedTuple

### GAME ###
class Game:
    players: list[Player]
    dealer: Dealer
    deck: Deck
    min_bet:int
    log: Log
    out_players: list[Player]

    def __init__(self, players:list[Player], deck:Deck=Deck(shuffle=True, num_decks=8), min_bet:int=15, hit_on_soft_17:bool=False) -> None:
        self.players = players
        self.out_players = []
        self.dealer = Dealer()
        self.deck = deck
        self.min_bet = min_bet
        self.hit_on_soft_17 = hit_on_soft_17
        self.log = Log()

    def hit_player(self,player:Player) -> bool:
        # if out of cards
        if (len(self.deck) == 0):
            self.deck.reset()
        card = self.deck.deal()
        return player.hit(card)

    def add_player(self,player:Player) -> None:
        self.players.append(player)

    def start(self) -> GameResults:
        self._reset()
        # init screen
        esc.enable_alt_buffer(); esc.cursor_to_top()
        try:
            self._check_player_chips()
            #
            if len(self.players) < 1:
                self._print_exit()
                return
            # start bet round
            self._start_bet_round()
            # start initial hit round
            self._start_init_hit_round()
            # start decision round
            self._start_player_decision_round()
            # start dealer hit round
            self._start_dealer_hit_round()
            # handle game end
            game_results = self._get_results()
            # print results
            self._print_game_state(dealer=self.dealer,game_results=game_results, )
            # play again?
            if self._is_play_again():
                self._restart_game()

        # end round, disable screen
        finally:
            esc.disable_alt_buffer()
            self._print_exit()
        
        return game_results

    def _check_player_chips(self):
        for player in self.players:
            if player.chips < self.min_bet:
                self.out_players.append(player)
        
        for player in self.out_players:
            if player in self.players:      
                self.players.remove(player)
                
    def _start_bet_round(self) -> None:
        for player in self.players:
            player.get_init_round_inputs(min_bet=self.min_bet)
            esc.erase_screen(); esc.cursor_to_top()
    
    def _start_init_hit_round(self) -> None:
        for _ in range(2):
            for player in self.players:
                if player.has_pseudos():
                    for pseudo in player.pseudos:
                        # pseudo.hit(Card("9"))
                        self.hit_player(player=pseudo)
                else:
                    # player.hit(Card("9"))
                    self.hit_player(player=player)

            self.hit_player(player=self.dealer)
            # self.dealer.hit(Card("9"))

    def _start_player_decision_round(self) -> None:
        self._print_game_state()
        self.log.add(f"starting")
        i = 0
        while i < len(self.players):
            self.log.add(f"i = {i}")
            player = self.players[i]
            i += 1
        # for player in self.players:
            if player.has_pseudos():
                j = 0
                while j < len(player.pseudos) :
                    self.log.add(f"j = {j}")
                    pseudo = player.pseudos[j]
                    j+=1
                    self._print_game_state()
                    if pseudo.has_blackjack():
                        self._handle_player_blackjack(player=pseudo)
                        continue
                    decision = self._get_player_decision(player=pseudo)
                    if decision in ["spl","split", "i","insurance", "hit","h"]:
                        j -= 1
                    self._handle_player_decision(player=pseudo, decision=decision)
                    if pseudo.is_bust():
                        j += 1
                    self._print_game_state()
            else:
                self._print_game_state(current_player=player)
                if player.has_blackjack():
                    self._handle_player_blackjack(player=player)
                    self._print_game_state()
                    continue
                decision = self._get_player_decision(player=player)
                self.log.add(decision)
                if decision in ["spl","split", "i","insurance", "hit","h"]:
                    i -= 1
                self._handle_player_decision(player=player, decision=decision)
                if player.is_bust():
                    i += 1
            self._print_game_state()
            
    def _handle_player_blackjack(self, player:Player) -> None:
        if player.has_blackjack():
            self.log.add(f"{player.name} has Black Jack!", "Green/italic/bold")
            if self.dealer.showing_ace():
                player_choice = esc.input(f"Insurance? [y/N]\n> ", input="Magenta",end="")
                if str.lower(player_choice) == "y":
                    player.place_insurance_bet()
            player.is_stayed = True

    def _get_player_decision(self, player:Player) -> str:
        keyclr = "Magenta"
        valid_inps = ["","Stay","s", "Hit", "h"]
        
        if player.hand.len() <= 2:
            if (player.is_pseudo and player.parent.chips > player.bet) or player.chips > player.bet:
                valid_inps += ["Double Down", "dd"]
            if player.can_split():
                valid_inps += ["Split", "spl"]

        if self.dealer.hand[0].rank == "Ace":
            if not player.has_insurance():
                valid_inps += ["Insurance","i"]
        
        esc.printf((player.name,"Cyan/underline/bold")," -> ", (" | ".join(list(map(lambda card: card.to_str(), player.hand))), "Cyan/bold"), end="")
        for i in range(1,len(valid_inps),2):
            post_str = ", " if i+2 != len(valid_inps) else ""
            print(f"{valid_inps[i]} (", end=""); esc.print(valid_inps[i+1], keyclr, end=""); print(")"+post_str, end="")

        print() 
        player_inp = esc.input("> ", input=keyclr, end="")

        if str.lower(player_inp) not in list(map(lambda s:str.lower(s), valid_inps)):
            esc.erase_prev(3)
            esc.print(f'"{player_inp}" not a valid input', "red")
            return self._get_player_decision(player=player)

        return player_inp
    
    def _handle_player_decision(self, player:Player, decision:str) -> None:
        decision = str.lower(decision)
        # STAY
        if decision in ["s","stay", ""]:
            self.log.add(f"{player.name} has Stayed", "Blue/italic")
            player.is_stayed = True
        # DOUBLE DOWN
        elif decision in ["dd", "double down"]:
            self._handle_player_double_down(player=player)
        # SPLIT
        elif decision in ["spl","split"]:
            player.split_hand()
        # HIT
        elif decision in ["hit","h"]:
            self.hit_player(player=player)
        # INSURANCE
        elif decision in ["i", "insurance"]:
            # esc.print(f"{player.name} chose Insurance... insurance bet set @ {round(player.bet/2)}", "Green")
            player.place_insurance_bet()

    def _handle_player_double_down(self, player:Player) -> None:
        self.log.add(f"{player.name} has doubled down", "Blue/italic")
        player.place_bet(bet_amount=player.bet)
        self.hit_player(player=player)
        player.is_stayed = True
        if player.is_bust():
            self.log.add(f"{player.name} has busted", "red/italic")

    def _start_dealer_hit_round(self) -> None:
        if self.hit_on_soft_17:
            while self.dealer.lowest_hand_value() < 17:
                self.hit_player(player=self.dealer)
        else:
            while self.dealer.hand_value() < 17:
                self.hit_player(player=self.dealer)
    
    def _get_results(self) -> GameResults:
        game_results = GameResults()
        for player in self.players:
            p_res_dict = {
                "player":player,
                "hands" : 0,
                "won" : 0,
                "busted" : 0,
                "pushed" : 0,
                "net" : 0
            }
            # multiple / split hands
            if player.has_pseudos():
                for pseudo in player.pseudos:
                    p_res_dict["hands"] += 1
                    if pseudo.has_insurance():
                        if self.dealer.has_blackjack():
                            player.chips += (2*pseudo.insurance)
                            p_res_dict["net"] += (pseudo.insurance)
                        else:
                            p_res_dict["net"] -= pseudo.insurance

                    if pseudo.is_bust():
                        p_res_dict["busted"] += 1
                        p_res_dict["net"] -= pseudo.bet

                    
                    elif self.dealer.is_bust():
                        p_res_dict["won"] += 1
                        player.chips += (pseudo.bet * 2)
                        p_res_dict["net"] += pseudo.bet
                    
                    elif self.dealer.has_blackjack():
                        if pseudo.has_blackjack():
                            p_res_dict["pushed"] += 1
                        else:
                            p_res_dict["net"] -= pseudo.bet

                    elif pseudo.has_blackjack():
                        p_res_dict["won"] += 1
                        player.chips += (pseudo.bet * 1.5)
                        p_res_dict["net"] += (pseudo.bet * .5)
                    
                    elif pseudo.hand_value() == self.dealer.hand_value():
                        p_res_dict["pushed"] += 1
                        player.chips += pseudo.bet
                    
                    elif pseudo.hand_value() > self.dealer.hand_value():
                        p_res_dict["won"] += 1
                        player.chips += (pseudo.bet * 2)
                        p_res_dict["net"] += (pseudo.bet)
                    
                    elif player.hand_value() < self.dealer.hand_value():
                        p_res_dict["net"] -= pseudo.bet
            # 1 hand
            else:
                p_res_dict["hands"] = 1

                if player.has_insurance():
                    if self.dealer.has_blackjack():
                        player.chips += (2*player.insurance)
                        p_res_dict["net"] += (player.insurance)
                    else:
                        p_res_dict["net"] -= player.insurance

                if player.is_bust():
                    p_res_dict["busted"] += 1
                    p_res_dict["net"] -= player.bet

                elif self.dealer.is_bust():
                    p_res_dict["won"] += 1
                    player.chips += (player.bet * 2)
                    p_res_dict["net"] += player.bet
                
                elif self.dealer.has_blackjack():
                    if player.has_blackjack():
                        p_res_dict["pushed"] += 1
                    else:
                        p_res_dict["net"] -= player.bet

                elif player.has_blackjack():
                    p_res_dict["won"] += 1
                    player.chips += (player.bet * 1.5)
                    p_res_dict["net"] += (player.bet * .5)
                                
                elif player.hand_value() == self.dealer.hand_value():
                    p_res_dict["pushed"] += 1
                    player.chips += player.bet
                
                elif player.hand_value() > self.dealer.hand_value():
                    p_res_dict["won"] += 1
                    player.chips += (player.bet * 2)
                    p_res_dict["net"] += (player.bet)
                
                elif player.hand_value() < self.dealer.hand_value():
                    p_res_dict["net"] -= player.bet

            p_results = PlayerResults(**p_res_dict)
            player.results.append(p_results)
            game_results.add(p_results)
        return game_results

    def _is_play_again(self) -> bool:
        return str.lower(input("Play again? (Y/n) \n> ")) != "n"
    
    def _reset(self) -> None:
        [player.reset() for player in self.players]
        self.dealer.reset()

    def _restart_game(self) -> GameResults:
        esc.erase_screen()
        self._reset()
        return self.start()

    def _get_player_max_name_len(self) -> int:
        max_len = len(self.dealer.name)
        for player in self.players:
            if player.has_pseudos():
                for pseudo in player.pseudos:
                    if len(pseudo.name) > max_len:
                        max_len = len(pseudo.name)
            else: 
                if len(player.name) > max_len:
                    max_len = len(player.name)
        return max_len

    def _print_game_state(self, current_player:Player=None, reset:bool=True, dealer:Dealer=None, game_results:GameResults=None) -> None:
        hide_dealer = (dealer == None)
        if reset:
            esc.erase_screen()
            esc.cursor_to_top()
        mxnmlen = self._get_player_max_name_len()
        # if current_player:
        #     esc.set("dim")
        self.dealer.print(max_name_len=mxnmlen, hidden=hide_dealer)
        print()
        for player in self.players:
            # if current_player and player != current_player:
            #     esc.set("dim")
            player.print(max_name_len=mxnmlen, dealer=dealer)
            print()
        if game_results:
            game_results.print()
        # self.log.print()
        print()

    def _print_exit(self) -> None:
        for player in self.players:
            esc.printf(
                (player.name,"Magenta"), " ended with ",(f"${player.chips}","Magenta"),
            )
        for player in self.out_players:
            esc.printf(
                (player.name,"Magenta"), " ended with ",(f"${player.chips}","Magenta"),
            )
        print()

    def _is_all_players_bust(self) -> bool:
        for player in self.players:
            if not player.is_bust():
                return False
        return True

    @staticmethod
    def create(players:list[str|tuple]) -> Game:
        players = [
            Player(name=player[0],strategy=player[1] if len(player) > 1 else Simple()) if isinstance(player,tuple) else Player(name=player)
            for player in players
        ]
        return Game(players=players)

### SIMULATION
class Simulation(Game):
    def __init__(self, players: list[Player], deck:Deck = Deck(shuffle=True, num_decks=8), min_bet:int = 15, hit_on_soft_17:bool=False) -> None:
        super().__init__(players, deck, min_bet, hit_on_soft_17)
    
    def run(self, n_times:int=1, print_sim:bool=False, wait:float=.01) -> SimulationResults:
        start_time = time()

        if print_sim:
            esc.enable_alt_buffer(); 
            esc.cursor_to_top()

        for i in range(n_times):    
            if len(self.players) < 1:
                break
            self._start(print_sim=print_sim)
            if print_sim:
                sleep(wait)
                esc.erase_screen()
                esc.cursor_to_top()

        time_elapsed = time() - start_time

        if print_sim:
            esc.disable_alt_buffer()
        
        results = SimulationResults(players=self.players+self.out_players, n_times=n_times, time_elapsed=time_elapsed)
        
        return results

    def _start(self, print_sim:bool=False) -> GameResults:
        self._reset()
        #
        self._check_player_chips()
        # get hand & bet amount
        self._handle_init_round_inputs()
        # handle init hit rounds
        self._start_init_hit_round()
        # handle decision rounds
        self._handle_decision_round()
        # handle dealer hit round
        self._start_dealer_hit_round()
        # print game state
        if print_sim:
            self._print_game_state(dealer=self.dealer, reset=False)
        
        self._handle_post_game_strat()

        results = self._get_results()
        if print_sim:
            results.print()

        return results
    
    def _handle_init_round_inputs(self) -> None:
        for player in self.players:
            n_hands = player.strategy.__decide_hands__(player=player)
            p_bet = player.strategy.__decide_bet__(player=player, min_bet=self.min_bet)
            if p_bet < self.min_bet:
                p_bet = self.min_bet
            if n_hands > 1:
                player._handle_mult_hands(hand_amount=n_hands, bet_amount=p_bet)
            else:
                player.place_bet(p_bet, min_bet=self.min_bet)
    
    def _handle_decision_round(self) -> None:
        self.log.add(f"starting")
        i = 0
        while i < len(self.players):
            player = self.players[i]
            i += 1
        # for player in self.players:
            if player.has_pseudos():
                j = 0
                while j < len(player.pseudos) :
                    pseudo = player.pseudos[j]
                    j+=1
                    if pseudo.has_blackjack():
                        self._handle_player_blackjack(player=pseudo)
                        continue

                    decision = self._get_player_decision(player=pseudo)
                    if decision in ["spl","split", "i","insurance", "hit","h"]:
                        j -= 1
                    
                    self._handle_player_decision(player=pseudo, decision=decision)
                    if pseudo.is_bust():
                        j += 1
            else:
                if player.has_blackjack():
                    self._handle_player_blackjack(player=player)
                    continue
        
                decision = self._get_player_decision(player=player)
                if decision in ["spl","split", "i","insurance", "hit","h"]:
                    i -= 1
                self._handle_player_decision(player=player, decision=decision)
                if player.is_bust():
                    i += 1
     
    def _handle_player_blackjack(self, player:Player) -> None:
        if player.has_blackjack():
            if self.dealer.showing_ace():
                decision = self._get_player_decision(player=player, choices=["insurance","stay"])
                self._handle_player_decision(player=player, decision=decision)
            player.is_stayed = True

    def _get_player_decision(self, player: Player, choices:list[str]=None) -> str:
        if not choices:
            choices = self._get_valid_choices(player=player)
        players = list(filter(lambda p: p != player, self.players))
        decision = player.strategy.__decide__(player=player, choices=choices, dealer=self.dealer, players=players)
        return decision

    def _get_valid_choices(self, player:Player) -> list[str]:
        valid_inps = ["","stay","hit"]
        
        if player.hand.len() <= 2:
            if (player.is_pseudo and player.parent.chips > player.bet) or player.chips > player.bet:
                valid_inps.append("double down")
            if player.can_split():
                valid_inps.append("split")

        if self.dealer.hand[0].rank == "Ace":
            if not player.has_insurance():
                valid_inps.append("insurance")
        
        return valid_inps
        
    def _handle_post_game_strat(self) -> None:
        for player in self.players:
            players = list(filter(lambda p: p != player, self.players))
            player.strategy.__after__(player=player, players=players, dealer=self.dealer)
### GAME RESULTS
class GameResults(list[PlayerResults]):
    def __init__(self) -> None:
        super().__init__()
    def add(self, player_res:PlayerResults) -> None:
        return self.append(player_res)
    def print(self) -> None:
        for result in self:
            # print(result.net)
            net = (f"+(${result.net})","Green") if result.net >= 0 else (f"-(${abs(result.net)})","Red")
            _and = " with " if result.pushed > 0 else ""
            pushed = (f"{result.pushed}","Magenta") if result.pushed > 0 else ""
            plshands = " hands pushed." if result.pushed > 0 else ""
            esc.printf(
                (f"{result.player.name} ","Magenta"), net, ": ",(f"{result.won}","Magenta"), "/",(f"{result.hands}","Magenta"), " hands won", _and, pushed, plshands
            )

### SIMULATION RESULTS
class SimulationResults(dict[str,PlayerSimulationResults]):
    def __init__(self, players:list[Player], n_times:int, time_elapsed:float=0.0) -> None:
        self.players = players
        self.n_times = n_times
        self.update()

    def update(self):
        for player in self.players:
            self[player.name] = PlayerSimulationResults(
                player=player,
                rounds=self.n_times,
                hands=sum(res.hands for res in player.results),
                won=sum(res.won for res in player.results),
                pushed=sum(res.pushed for res in player.results),
                busted=sum(res.busted for res in player.results),
                net=(player.chips - player.init_chips),
                win_rate=(sum(res.won for res in player.results) / sum(res.hands for res in player.results))
            )
    
    def print(self):
        for name in self:
            res = self[name]
            rate_sty = "Red" if res.win_rate < .5 else "Green"
            net_sty = "Red" if res.net < 0 else "Green"
            esc.printf(
                (f"{name}", "Magenta"), " won ",(f"{round(res.win_rate * 100, 2)}%", rate_sty), " of the time with net chip earnings = ",(f"{res.net}",net_sty),
            )
            print()

### LOG
class Log(list[tuple[str,str]]):
    def __init__(self) -> None:
        pass

    def add(self, log_item:str, style:str=""):
        self.append((log_item, style))
        
    def delete(self, key:str):
        for i in range(len(self)):
            if self[i][0] == key:
                del self[i]
        
    def print(self):
        for item in self:
            esc.print('~ ' + item[0], item[1])
